# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Version string parsing for the feature-check Python library."""

from __future__ import annotations

from typing import NamedTuple


class VersionComponent(NamedTuple):
    """Represent a single version component: a numerical part and a freeform string one."""

    num: int | None
    rest: str

    def __str__(self) -> str:
        """Provide a string representation of the version component."""
        return (str(self.num) if self.num is not None else "") + self.rest

    def cmp(self, other: object) -> int:  # noqa: C901,PLR0911,PLR0912
        """Compare two components, return None if they are equal."""
        if not isinstance(other, VersionComponent):
            raise NotImplementedError(repr(other))

        if self.num is not None:
            if other.num is not None:
                if self.num < other.num:
                    return -1
                if self.num > other.num:
                    return 1

                if self.rest is not None:
                    if other.rest is not None:
                        if self.rest < other.rest:
                            return -1
                        if self.rest > other.rest:
                            return 1
                        return 0

                    return 1

                if other.rest is not None:
                    return -1

                return 0

            return 1

        if other.num is not None:
            return -1

        if self.rest is not None:
            if other.rest is not None:
                if self.rest < other.rest:
                    return -1
                if self.rest > other.rest:
                    return 1
                return 0

            return -1

        return 0

    def __lt__(self, other: object) -> bool:
        """Check whether this version component is less than the other one."""
        return self.cmp(other) < 0

    def __le__(self, other: object) -> bool:
        """Check whether this version component is less than or equal to the other one."""
        return self.cmp(other) <= 0

    def __eq__(self, other: object) -> bool:
        """Check whether this version component is equal to the other one."""
        return self.cmp(other) == 0

    def __ne__(self, other: object) -> bool:
        """Check whether this version component is not equal to the other one."""
        return self.cmp(other) != 0

    def __ge__(self, other: object) -> bool:
        """Check whether this version component is greater than or equal to the other one."""
        return self.cmp(other) >= 0

    def __gt__(self, other: object) -> bool:
        """Check whether this version component is greater than the other one."""
        return self.cmp(other) > 0

    def __hash__(self) -> int:
        """Return an integer key for hashable collections."""
        return hash((self.num, self.rest))


class Version(NamedTuple):
    """A version string: many components, possibly other attributes."""

    value: str
    components: list[VersionComponent]


def _version_compare_split_empty(
    spl_a: list[VersionComponent],
    spl_b: list[VersionComponent],
) -> int | None:
    """Check if any of the split version numbers is empty."""
    if not spl_a:
        if not spl_b:
            return 0
        if spl_b[0].num is None:
            return 1
        return -1
    if not spl_b:
        if spl_a[0].num is None:
            return -1
        return 1

    return None


def _version_compare_split(spl_a: list[VersionComponent], spl_b: list[VersionComponent]) -> int:
    """Compare two version numbers already split into component lists.

    Returns -1, 0, or 1 for the first version being less than, equal to,
    or greater than the second one.
    """
    res = _version_compare_split_empty(spl_a, spl_b)
    if res is not None:
        return res

    (comp_a, comp_b) = (spl_a.pop(0), spl_b.pop(0))
    res = comp_a.cmp(comp_b)
    if res != 0:
        return res

    return _version_compare_split(spl_a, spl_b)


def version_compare(ver_a: Version, ver_b: Version) -> int:
    """Compare two version strings.

    Returns -1, 0, or 1 for the first version being less than, equal to,
    or greater than the second one.
    """
    return _version_compare_split(ver_a.components, ver_b.components)



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Base module for the feature-check Python library.

This module imports the `obtain_features()`, `parse_expr()`, and
`parse_version()` functions from the `feature_check.obtain`,
`feature_check.expr`, and `feature_check.version` submodules;
see their documentation for details.
"""

# isort: skip_file

from .defs import VERSION
from .expr import ResultBool, ResultVersion
from .obtain import obtain_features
from .parser import parse_expr, parse_version
from .version import Version, VersionComponent

__all__ = [
    "VERSION",
    "ResultBool",
    "ResultVersion",
    "Version",
    "VersionComponent",
    "obtain_features",
    "parse_expr",
    "parse_version",
]



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Parse feature names, versions, and simple expressions."""

from __future__ import annotations

from typing import Final

import pyparsing as pyp

from . import defs
from . import expr as fexpr
from . import version as fver


_p_comp_num: Final = pyp.Word(pyp.srange("[0-9]"))

_p_comp_rest: Final = pyp.Word(pyp.srange("[A-Za-z~-]"), pyp.srange("[0-9A-Za-z~-]"))

_p_version_comp: Final = (
    _p_comp_num.set_results_name("num") + pyp.Opt(_p_comp_rest.set_results_name("rest_first"))
) | _p_comp_rest.set_results_name("rest_only")

_p_version: Final = _p_version_comp + pyp.ZeroOrMore(pyp.Char(".").suppress() + _p_version_comp)

_p_feature: Final = pyp.Word(pyp.srange("[A-Za-z0-9_-]"))

_p_op_sign: Final = pyp.one_of(["<=", "<", "=", ">=", ">"])

_p_op_word: Final = pyp.one_of(["le", "lt", "eq", "ge", "gt"])

_p_op_sign_and_value: Final = (
    pyp.Opt(pyp.White()).suppress() + _p_op_sign + pyp.Opt(pyp.White()).suppress() + _p_version
)

_p_op_word_and_value: Final = (
    pyp.White().suppress() + _p_op_word + pyp.White().suppress() + _p_version
)

_p_expr: Final = _p_feature + pyp.Opt(_p_op_sign_and_value | _p_op_word_and_value)

_p_feature_version: Final = _p_feature + pyp.Opt(pyp.Literal("=").suppress() + _p_version)

_p_features_line: Final = _p_feature_version + pyp.ZeroOrMore(
    pyp.White().suppress() + _p_feature_version,
)


@_p_version_comp.set_parse_action
def _process_version_comp(tokens: pyp.ParseResults) -> fver.VersionComponent:
    """Build a `VersionComponent` object out of the numeric and string parts."""
    tok_dict: Final[dict[str, str]] = tokens.as_dict()
    num_str: Final = tok_dict.get("num")
    rest: Final = tok_dict.get("rest_first", tok_dict.get("rest_only", ""))
    return fver.VersionComponent(num=int(num_str) if num_str is not None else None, rest=rest)


@_p_version.set_parse_action
def _process_version(tokens: pyp.ParseResults) -> fver.Version:
    """Build a `Version` object out of the version components."""
    res: Final = tokens.as_list()
    if not all(isinstance(comp, fver.VersionComponent) for comp in res):
        raise ParseError(2, f"Weird version parse result: {res!r}")

    return fver.Version(value=".".join(str(comp) for comp in res), components=res)


@_p_op_sign.set_parse_action
def _parse_op_sign(tokens: pyp.ParseResults) -> fexpr.BoolOp:
    """Parse a boolean operation written as a sign ("<", ">=", etc)."""
    return fexpr.OPS[tokens[0]]


@_p_op_word.set_parse_action
def _parse_op_word(tokens: pyp.ParseResults) -> fexpr.BoolOp:
    """Parse a boolean operation written as a sign ("lt", "ge", etc)."""
    return fexpr.OPS[tokens[0]]


@_p_expr.set_parse_action
def _parse_expr(tokens: pyp.ParseResults) -> defs.Mode:
    """Build a `Mode` out of a single feature name or a simple expression."""
    res: Final = tokens.as_list()
    match res:
        case [feature_name] if isinstance(feature_name, str):
            return defs.ModeSingle(feature=feature_name, ast=fexpr.ExprFeature(feature_name))

        case [feature_name, cmp_op, ver] if isinstance(feature_name, str) and isinstance(
            cmp_op,
            fexpr.BoolOp,
        ) and isinstance(ver, fver.Version):
            return defs.ModeSimple(
                ast=fexpr.ExprOp(
                    op=cmp_op,
                    args=[fexpr.ExprFeature(feature_name), fexpr.ExprVersion(ver)],
                ),
            )

    raise ParseError(2, f"Weird expr parse results: {res!r}")


@_p_feature_version.set_parse_action
def _parse_feature_version(tokens: pyp.ParseResults) -> tuple[str, fver.Version]:
    """Parse a feature name and a version, defaulting to "1.0"."""
    res: Final = tokens.as_list()
    match res:
        case [feature] if isinstance(feature, str):
            return (feature, parse_version("1.0"))

        case [feature, ver] if isinstance(feature, str) and isinstance(ver, fver.Version):
            return (feature, ver)

    raise ParseError(2, f"Weird feature/version parse results: {res!r}")


@_p_features_line.set_parse_action
def _parse_features_line(tokens: pyp.ParseResults) -> dict[str, fver.Version]:
    """Build a features dictionary out of the parsed name/version tuples."""

    # We lie a little bit, or mypy won't be happy.
    def validate_pair(pair: tuple[str, fver.Version]) -> tuple[str, fver.Version]:
        """Validate a single parsed result."""
        match pair:
            case (name, value) if isinstance(name, str) and isinstance(value, fver.Version):
                return name, value

        raise ParseError(2, f"Weird features line parse result: {pair!r}")

    res: Final = tokens.as_list()
    return dict(validate_pair(pair) for pair in res)


_p_version_complete: Final = _p_version.leave_whitespace()

_p_expr_complete: Final = _p_expr.leave_whitespace()

_p_features_line_complete: Final = _p_features_line.leave_whitespace()


class ParseError(defs.FCError):
    """An error that occurred while parsing the expression."""


def parse_version(value: str) -> fver.Version:
    """Parse a version string into a `Version` object."""
    res: Final = _p_version_complete.parse_string(value).as_list()
    match res:
        case [ver] if isinstance(ver, fver.Version) and ver.value == value:
            return ver

        case [ver]:
            raise ParseError(
                2,
                f"Could not parse the whole of {value!r} as a version string "
                f"(parsed {ver!r} from {res!r})",
            )

    raise ParseError(2, f"Could not parse {value!r} as a version string (result: {res!r})")


def parse_expr(expr: str) -> defs.Mode:
    """Parse a simple "feature-name op version" expression.

    If the expression is valid, return an `Expr` object corresponding
    to the specified check.  Use this object's `evaluate()` method and
    pass a features dictionary as returned by the `obtain_features()`
    function to get a `Result` object; for simple expressions it will be
    a `ResultBool` object with a boolean `value` member.

        from feature_check import expr as fexpr
        from feature_check import obtain as fobtain

        data = fobtain.obtain_features("timelimit");
        expr = fexpr.parse_simple("subsecond > 0")
        print(expr.evaluate(data).value)
    """
    res: Final = _p_expr_complete.parse_string(expr).as_list()
    match res:
        case [mode] if isinstance(mode, defs.Mode):
            return mode

    raise ParseError(2, f"Could not parse {expr!r} as an expression (results: {res!r})")


def parse_features_line(features_line: str) -> dict[str, fver.Version]:
    """Parse the features list, default to version "1.0"."""
    try:
        res: Final = _p_features_line_complete.parse_string(features_line).as_list()
    except pyp.exceptions.ParseBaseException as err:
        raise ParseError(2, f"Could not parse {features_line!r} as a features line: {err}") from err
    match res:
        case [features] if isinstance(features, dict):
            return features

    raise ParseError(
        2,
        f"Could not parse {features_line!r} as a features line (results: {res!r})",
    )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""A set of Hypothesis-based correctness tests."""

from __future__ import annotations

import contextlib
import subprocess  # noqa: S404
import typing
from unittest import mock

import hypothesis as hyp
import hypothesis.strategies as strat

from feature_check import defs as fdefs
from feature_check import obtain as fobtain
from feature_check import parser as fparser


if typing.TYPE_CHECKING:
    from typing import Final


ALPHABET_NAME: Final = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"
ALPHABET_VER_FIRST: Final = "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
ALPHABET_VER_REST: Final = f"0{ALPHABET_VER_FIRST}"

STRAT_COMPONENT: Final = strat.tuples(
    strat.text(ALPHABET_VER_FIRST, min_size=1, max_size=1),
    strat.text(ALPHABET_VER_REST),
)

STRAT_FEATURES: Final = strat.lists(
    strat.tuples(strat.text(ALPHABET_NAME, min_size=1), strat.lists(STRAT_COMPONENT, min_size=1)),
    min_size=1,
)


def build_features(pairs: list[tuple[str, list[tuple[str, str]]]]) -> list[tuple[str, str]]:
    """Build a list of (name, version) pairs out of the feature components."""
    return [(name, ".".join(f"{first}{rest}" for first, rest in comps)) for name, comps in pairs]


@hyp.given(line=strat.text())
@hyp.example("")
def test_parse_exception(line: str) -> None:
    """Make sure the parser returns either success or our own exception."""
    with contextlib.suppress(fdefs.FCError):
        fparser.parse_features_line(line)


@hyp.given(pairs=STRAT_FEATURES)
def test_parse_pairs(pairs: list[tuple[str, list[tuple[str, str]]]]) -> None:
    """Make sure the parser returns the expected dictionary of feature names and versions."""
    paired: Final = build_features(pairs)
    exp: Final = dict(paired)
    line: Final = " ".join(f"{name}={ver}" for name, ver in paired)
    features: Final = fparser.parse_features_line(line)
    res: Final = {name: ver.value for name, ver in features.items()}
    assert res == exp


@hyp.given(
    program=strat.text(min_size=1),
    option=strat.text(min_size=1),
    prefix=strat.text(ALPHABET_NAME),
    pairs=STRAT_FEATURES,
)
@hyp.example("0", "0", "", [("A", [("1", "")])])
def test_parse_features_line(
    program: str,
    option: str,
    prefix: str,
    pairs: list[tuple[str, list[tuple[str, str]]]],
) -> None:
    """Make sure a full line can be parsed."""
    paired: Final = build_features(pairs)
    exp: Final = dict(paired)
    line: Final = prefix + " ".join(f"{name}={ver}" for name, ver in paired)

    def mock_subprocess_run(
        args: list[str],
        *,
        capture_output: bool = False,
        check: bool = True,
        stdin: str | None = "oof?",
    ) -> subprocess.CompletedProcess[bytes]:
        """Mock the invocation of an external program."""
        assert args == [program, option]
        assert capture_output
        assert not check
        assert stdin is None
        return subprocess.CompletedProcess(
            args=args,
            returncode=0,
            stdout=f"{line}\n".encode(),
            stderr=b"",
        )

    with mock.patch("subprocess.run", new=mock_subprocess_run):
        features: Final = fobtain.obtain_features(program, option, prefix)

    res: Final = {name: ver.value for name, ver in features.items()}
    assert res == exp



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Query a program's list of features."""

from __future__ import annotations

import dataclasses
import sys
from typing import Final

import click


try:
    import simplejson as js
except ImportError:
    import json as js  # type: ignore[no-redef]

from . import defs
from . import expr as fexpr
from . import obtain
from . import parser as fparser
from . import version as fver


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for this program."""

    args: list[str]
    display_version: bool
    features_prefix: str
    option_name: str
    output_format: str

    program: str = "(unknown)"


def arg_version(_ctx: click.Context, _self: click.Parameter, value: bool) -> bool:  # noqa: FBT001
    """Display program version information."""
    if not value:
        return value

    print(f"feature-check {defs.VERSION}")
    sys.exit(0)


def arg_features(_ctx: click.Context, _self: click.Parameter, value: bool) -> bool:  # noqa: FBT001
    """Display program features information."""
    if not value:
        return value

    print(f"{defs.DEFAULT_PREFIX}feature-check={defs.VERSION} single=1.0 list=1.0 simple=1.0")
    sys.exit(0)


def output_tsv(data: dict[str, fver.Version]) -> None:
    """List the obtained features as tab-separated name/value pairs."""
    for feature in sorted(data.keys()):
        print(f"{feature}\t{data[feature].value}")


def output_json(data: dict[str, fver.Version]) -> None:
    """List the obtained features as a JSON object."""
    print(
        js.dumps(
            {name: value.value for name, value in data.items()},
            sort_keys=True,
            indent=2,
        ),
    )


OUTPUT = {"tsv": output_tsv, "json": output_json}


def process(mode: defs.Mode, cfg: Config, data: dict[str, fver.Version]) -> None:
    """Perform the requested feature-check operation."""
    match mode:
        case defs.ModeList():
            OUTPUT[cfg.output_format](data)

        case defs.ModeSingle(feature, _):
            if feature in data:
                if cfg.display_version:
                    print(data[feature].value)
                sys.exit(0)
            else:
                sys.exit(1)

        case defs.ModeSimple(ast):
            res: Final = ast.evaluate(data)
            if not isinstance(res, fexpr.ResultBool):
                sys.exit(f"Internal error: did not expect a {type(res).__name__} object")
            sys.exit(not res.value)

        case _:
            sys.exit(f"Internal error: process(mode={mode!r}, cfg={cfg!r}, data={data!r}")


@click.command(name="feature-check")
@click.help_option("--help", "-h")
@click.option(
    "--version",
    "-V",
    is_flag=True,
    is_eager=True,
    callback=arg_version,
    help="display program version information and exit",
)
@click.option(
    "--features",
    is_flag=True,
    is_eager=True,
    callback=arg_features,
    help="display supported features and exit",
)
@click.option(
    "-v",
    "--display-version",
    is_flag=True,
    help="display the feature version",
)
@click.option(
    "-P",
    "--features-prefix",
    type=str,
    default=defs.DEFAULT_PREFIX,
    help="the features prefix in the program output",
)
@click.option(
    "-l",
    "--list",
    "opt_list",
    is_flag=True,
    help="list the features supported by a program",
)
@click.option(
    "-O",
    "--option-name",
    type=str,
    default=defs.DEFAULT_OPTION,
    help="the query-features option to pass",
)
@click.option(
    "-o",
    "--output-format",
    type=click.Choice(sorted(OUTPUT.keys())),
    default=defs.DEFAULT_OUTPUT_FMT,
    help="specify the output format for the list",
)
@click.argument("program", type=str)
@click.argument("feature", type=str, required=False)
@click.argument("op", type=str, required=False)
@click.argument("feature_version", type=str, required=False, metavar="VERSION")
@click.pass_context
def main(  # noqa: PLR0913  # yep, we accept a lot of command-line options
    ctx: click.Context,
    *,
    program: str,
    feature: str | None,
    op: str | None,
    display_version: bool,
    feature_version: str | None,
    features: bool,
    features_prefix: str,
    opt_list: bool,
    option_name: str,
    output_format: str,
    version: bool,
) -> None:
    """Parse command-line arguments, do things."""

    def parse_mode() -> defs.Mode:
        """Determine the mode of operation."""
        if opt_list:
            return defs.ModeList()

        if not cfg.args:
            ctx.fail("No feature specified")
        expr: Final = " ".join(cfg.args)
        try:
            return fparser.parse_expr(expr)
        except fparser.ParseError:
            ctx.fail("Only querying a single feature supported so far")

    if features or version:
        ctx.fail(f"How did we get to main() with {features=!r} and {version=!r}?")

    cfg: Final = Config(
        program=program,
        args=[value for value in [feature, op, feature_version] if value is not None],
        display_version=display_version,
        features_prefix=features_prefix,
        option_name=option_name,
        output_format=output_format,
    )

    mode: Final = parse_mode()

    try:
        data: Final = obtain.obtain_features(cfg.program, cfg.option_name, cfg.features_prefix)
    except obtain.ObtainError as exc:
        sys.exit(exc.code)

    process(mode, cfg, data)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Constant definitions for the feature-check Python library."""

from __future__ import annotations

import abc
import dataclasses
import typing


if typing.TYPE_CHECKING:
    from . import version as fver


DEFAULT_OPTION = "--features"
"""The default command-line option to pass to a program to query for supported features."""

DEFAULT_PREFIX = "Features: "
"""The default prefix of the program's features output line."""

DEFAULT_OUTPUT_FMT = "tsv"
"""The default output format for the `feature-check` command-line tool."""

VERSION = "2.2.0"
"""The feature-check library version, SemVer-style."""

VERSION_STRING = VERSION


class FCError(Exception):
    """A base class for errors in handling a feature-check request."""

    def __init__(self, code: int, msg: str) -> None:
        """Initialize an error object."""
        super().__init__(msg)
        self._code = code
        self._msg = msg

    @property
    def code(self) -> int:
        """Return the numeric error code."""
        return self._code

    @property
    def message(self) -> str:
        """Return a human-readable error message."""
        return self._msg


@dataclasses.dataclass(frozen=True)
class Result:
    """The base class for an expression result."""


@dataclasses.dataclass(frozen=True)
class Expr(abc.ABC):
    """The (pretty much abstract) base class for an expression."""

    @abc.abstractmethod
    def evaluate(self, data: dict[str, fver.Version]) -> Result:
        """Evaluate the expression and return a Result object.

        Overridden in actual expression classes.
        """
        raise NotImplementedError(
            f"{type(self).__name__}.evaluate() must be overridden",  # noqa: EM102
        )


@dataclasses.dataclass(frozen=True)
class Mode:
    """Base class for the feature-check operating modes."""


@dataclasses.dataclass(frozen=True)
class ModeList(Mode):
    """List the features supported by the program."""


@dataclasses.dataclass(frozen=True)
class ModeSingle(Mode):
    """Query for the presence or the version of a single feature."""

    feature: str
    ast: Expr


@dataclasses.dataclass(frozen=True)
class ModeSimple(Mode):
    """Verify whether a simple 'feature op version' expression holds true."""

    ast: Expr



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Query a program for the list of features that it supports."""

from __future__ import annotations

import subprocess  # noqa: S404
from typing import Final

from . import defs
from . import parser as fparser
from . import version as fver


class ObtainError(defs.FCError):
    """A base class for errors in obtaining the program's features."""


class ObtainExecError(ObtainError):
    """An error that occurred while executing the queried program."""

    def __init__(self, err: str) -> None:
        """Initialize an error object."""
        super().__init__(1, err)


class ObtainNoFeaturesSupportError(ObtainExecError):
    """The program does not seem to support the "--features" option."""

    def __init__(self, program: str, option: str) -> None:
        """Store the program name."""
        super().__init__(
            f"The {program} program does not seem to support "
            f"the {option} option for querying features",
        )


class ObtainNoFeaturesError(ObtainError):
    """An error that occurred while looking for the features line."""

    def __init__(self, program: str, option: str, prefix: str) -> None:
        """Initialize an error object."""
        super().__init__(
            2,
            f"The '{program} {option}' output did not contain a single '{prefix}' line",
        )


def obtain_features(
    program: str,
    option: str = defs.DEFAULT_OPTION,
    prefix: str = defs.DEFAULT_PREFIX,
) -> dict[str, fver.Version]:
    """Execute the specified program and get its list of features.

    The program is run with the specified query option (default:
    "--features") and its output is examined for a line starting with
    the specified prefix (default: "Features: ").  The rest of the line
    is parsed as a whitespace-separated list of either feature names or
    "name=version" pairs.  The function returns a dictionary of the features
    obtained with their versions (or "1.0" if only a feature name was found
    in the program's output).

        import feature_check

        data = feature_check.obtain_features("timelimit")
        print(data.get("subsecond", "not supported"))

    For programs that need a different command-line option to list features:

        import feature_check

        print("SSL" in feature_check.obtain_features("curl",
                                                     option="--version"))
    """

    def try_run() -> list[str]:
        """Query a program about its features, return the output lines."""
        res: Final = subprocess.run(  # noqa: S603
            [program, option],
            capture_output=True,
            check=False,
            stdin=None,
        )
        if res.returncode != 0 or res.stderr.decode():
            # It does not support '--features', does it?
            raise ObtainNoFeaturesSupportError(program, option)

        return res.stdout.decode().split("\n")

    try:
        lines: Final = try_run()
    except ObtainExecError:
        raise
    # Yes, we do want to convert any error into an ObtainExecError one
    except Exception as exc:
        # Something went wrong in the --features processing
        raise ObtainExecError(str(exc)) from exc

    matching: Final = (
        [
            rest
            for line, rest in ((line, line.removeprefix(prefix)) for line in lines)
            if rest != line
        ]
        if prefix
        else [line for line in lines if line]
    )
    if len(matching) != 1:
        raise ObtainNoFeaturesError(program, option, prefix)

    return fparser.parse_features_line(matching[0])



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Expression evaluation for the feature-check Python library."""

from __future__ import annotations

import dataclasses
from collections.abc import Callable
from typing import Final

from . import defs
from . import version as fver


BoolOpFunction = Callable[[list[defs.Result]], bool]


@dataclasses.dataclass(frozen=True)
class ResultBool(defs.Result):
    """A boolean result of an expression; the "value" member is boolean."""

    value: bool

    def __str__(self) -> str:
        """Provide a human-readable representation of the calculation result."""
        return f"ResultBool: {self.value}"


@dataclasses.dataclass(frozen=True)
class ResultVersion(defs.Result):
    """A version number as a result of an expression.

    The "value" member is the version number string.
    """

    value: fver.Version

    def __str__(self) -> str:
        """Provide a human-readable representation of the calculation result."""
        return f"ResultVersion: {self.value.value}"


@dataclasses.dataclass(frozen=True)
class ExprFeature(defs.Expr):
    """An expression that returns a program feature name as a string."""

    name: str

    def evaluate(self, data: dict[str, fver.Version]) -> ResultVersion:
        """Look up the feature, return the result in a ResultVersion object."""
        return ResultVersion(value=data[self.name])


@dataclasses.dataclass(frozen=True)
class ExprVersion(defs.Expr):
    """An expression that returns a version number for a feature."""

    value: fver.Version

    def evaluate(self, _data: dict[str, fver.Version]) -> ResultVersion:
        """Return the version number as a ResultVersion object."""
        return ResultVersion(value=self.value)


@dataclasses.dataclass(frozen=True)
class BoolOp:
    """A two-argument boolean operation."""

    args: list[type[defs.Result]]
    action: BoolOpFunction


def _def_op_bool_ver(check: Callable[[int], bool]) -> BoolOpFunction:
    def _op_bool_ver(args: list[defs.Result]) -> bool:
        """Check whether the arguments are in the expected relation."""
        match args:
            case [left, right] if isinstance(left, ResultVersion) and isinstance(
                right,
                ResultVersion,
            ):
                return check(fver.version_compare(left.value, right.value))

        raise TypeError(args)

    return _op_bool_ver


NAMED_OPS = {
    "lt": BoolOp(
        args=[ResultVersion, ResultVersion],
        action=_def_op_bool_ver(lambda res: res < 0),
    ),
    "le": BoolOp(
        args=[ResultVersion, ResultVersion],
        action=_def_op_bool_ver(lambda res: res <= 0),
    ),
    "eq": BoolOp(
        args=[ResultVersion, ResultVersion],
        action=_def_op_bool_ver(lambda res: res == 0),
    ),
    "ge": BoolOp(
        args=[ResultVersion, ResultVersion],
        action=_def_op_bool_ver(lambda res: res >= 0),
    ),
    "gt": BoolOp(
        args=[ResultVersion, ResultVersion],
        action=_def_op_bool_ver(lambda res: res > 0),
    ),
}

SYNONYMS = {"<": "lt", "<=": "le", "=": "eq", ">=": "ge", ">": "gt"}

OPS = dict(
    list(NAMED_OPS.items()) + [(name, NAMED_OPS[value]) for name, value in SYNONYMS.items()],
)


@dataclasses.dataclass(frozen=True)
class ExprOp(defs.Expr):
    """A two-argument operation expression."""

    op: BoolOp
    args: list[defs.Expr]

    def __post_init__(self) -> None:
        """Validate the passed arguments."""
        if len(self.args) != len(self.op.args):
            raise ValueError((self.args, self.op.args))

    def evaluate(self, data: dict[str, fver.Version]) -> ResultBool:
        """Evaluate the expression over the specified data."""
        args: Final = [expr.evaluate(data) for expr in self.args]

        for idx, value in enumerate(args):
            if not isinstance(value, self.op.args[idx]):
                raise TypeError((idx, self.op.args[idx], value))

        return ResultBool(value=self.op.action(args))



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Test the version comparison functions."""

from __future__ import annotations

from typing import Final

import pytest

from feature_check import defs
from feature_check import expr as fexpr
from feature_check import parser as fparser
from feature_check import version as fver

from . import data


def parsed_features() -> dict[str, fver.Version]:
    """Parse the versions of the features used in the test."""
    return {name: fparser.parse_version(value) for name, value in data.FEATURES.items()}


def do_test_compare(var: str, op_name: str, right: str, *, expected: bool) -> None:
    """Test the comparison functions."""
    feature: Final = f"{var} {op_name} {right}"
    mode: Final = fparser.parse_expr(feature)
    assert isinstance(mode, defs.ModeSimple)
    expr: Final = mode.ast
    assert isinstance(expr, fexpr.ExprOp)
    assert len(expr.args) == 2  # noqa: PLR2004
    assert isinstance(expr.args[0], fexpr.ExprFeature)
    assert isinstance(expr.args[1], fexpr.ExprVersion)

    res: Final = expr.evaluate(parsed_features())
    assert isinstance(res, fexpr.ResultBool)
    assert res.value == expected


@pytest.mark.parametrize(("var", "op_name", "right", "expected"), data.COMPARE)
def test_compare(*, var: str, op_name: str, right: str, expected: bool) -> None:
    """Test the comparison functions with word operands."""
    return do_test_compare(var, op_name, right, expected=expected)


@pytest.mark.parametrize(("var", "op_name", "right", "expected"), data.COMPARE)
def test_synonyms(*, var: str, op_name: str, right: str, expected: bool) -> None:
    """Test the comparison functions with word operands."""
    return do_test_compare(var, data.SYNONYMS[op_name], right, expected=expected)


@pytest.mark.parametrize("var", (item[0] for item in data.COMPARE))
def test_single(*, var: str) -> None:
    """Test obtaining the version of a single feature."""
    mode: Final = fparser.parse_expr(var)
    assert isinstance(mode, defs.ModeSingle)
    assert mode.feature == var
    expr: Final = mode.ast
    assert isinstance(expr, fexpr.ExprFeature)
    assert expr.name == var

    res: Final = expr.evaluate(parsed_features())
    assert isinstance(res, fexpr.ResultVersion)
    ver: Final = res.value
    assert ver.value
    assert ver.components
